/* This file is part of mailfrom filter.
   Copyright (C) 2006 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdlib.h>
#include "mailfrom.h"

/* |hash_size| defines a sequence of symbol table sizes. These are prime
   numbers, the distance between each pair of them grows exponentially,
   starting from 64. Hardly someone will need more than 16411 symbols, and
   even if someone will, it is easy enough to add more numbers to the
   sequence. */

static unsigned int hash_size[] = {
    37,   101,  229,  487, 1009, 2039, 4091, 8191, 16411
};

/* |max_rehash| keeps the number of entries in |hash_size| table. */
static unsigned int max_rehash = sizeof (hash_size) / sizeof (hash_size[0]);

/* |hash_num| is the index to |hash_size| table. Current size of the 
   |symtab| table is thus |hash_size[hash_num]| */

static unsigned int hash_num;  

struct symtab {
	int state;     /* Entry state */
	union {
		char *name;               /* Each of the structures below
					     must begin with char*: */
		struct function fun;      /* if state == SYM_FUNC */
		struct builtin builtin;   /* if state == SYM_BUILTIN */
		struct variable variable; /* if state == SYM_VARIABLE */
		struct db_format db;      /* if state == SYM_DB_FORMAT */
	} v;
};

static struct symtab *symtable; /* Table of symbols */

static struct symtab *lookup_or_install(int type, const char *name,
					int install);

static unsigned
hash(const char *name)
{
	unsigned i;
	
	for (i = 0; *name; name++) {
		i <<= 1;
		i ^= *(unsigned char*) name;
	}
	return i % hash_size[hash_num];
}

/* Allocate |n| bytes of memory. If allocation fails, bail out and
   abort. */

void *
emalloc(size_t n)          
{
	void *p;
	
	p = malloc(n);
	if (p == 0) {
		parse_error("not enough memory");
		abort();
	}
	return p;
}

/* |rehash| is called when the symbol table becomes full. Its purpose is
   to select new table size, allocate storage for the resized table, install
   to it all data stored so far and reclaim the unused memory.

   |rehash| returns |0| on success and |1| otherwise. */
     
static int
rehash()
{
	struct symtab *old_symtab = symtable;
	struct symtab *sp;
	unsigned int i;
	
	if (++hash_num >= max_rehash) {
		return 1;
	}
	
	symtable = calloc(hash_size[hash_num], sizeof(symtable[0]));
	if (!symtable) {
		parse_error("not enough memory");
		abort();
	}
	if (old_symtab) {
		for (i = 0; i < hash_size[hash_num - 1]; i++) {
			if (old_symtab[i].state != SYM_UNDEF) {
				sp = lookup_or_install(old_symtab[i].state,
						       old_symtab[i].v.name, 1);
				sp->state = old_symtab[i].state;
				memcpy(&sp->v, &old_symtab[i].v, sizeof sp->v);
			}
		}
		free(old_symtab);
	}
	return 0;
}

/* Look up symbol |name| in the table. If not found and |install| is not 0, 
   insert the symbol into the table.
   
   Basically it uses algorithm L from TAOCP 6.4, except that if the insertion
   fails the table is rehashed and the attempt is retried. The process 
   continues until the symbol is successfully inserted or the table reaches
   maximum allowed size. */

static struct symtab *
lookup_or_install(int state, const char *name, int install)
{
	unsigned i, pos;

	if (!symtable) {
		if (install) {
			if (rehash())
				return NULL;
		} else
			return NULL;
	}
	
	pos = hash(name);
	
	for (i = pos; symtable[i].state != SYM_UNDEF;) {
		if ((state == SYM_UNDEF || symtable[i].state == state)
		    && strcmp(symtable[i].v.name, name) == 0)
			return &symtable[i];
		if (++i >= hash_size[hash_num])
			i = 0;
		if (i == pos)
			break;
	}
	
	if (!install || state == SYM_UNDEF)
		return NULL;
	
	if (symtable[i].state == SYM_UNDEF)
		return &symtable[i];
	
	if (rehash())
		return NULL;
	
	return lookup_or_install(state, name, install);
}

int
symbol_enumerate(int state, symbol_enumerator_t fun, void *data)
{
	unsigned i;
	
	for (i = 0; i < hash_size[hash_num]; i++) {
		if (state == SYM_UNDEF || symtable[i].state == state) {
			int rc = fun(&symtable[i].v, data);
			if (rc)
				return rc;
		}
	}
	return 0;
}

	  
void
va_builtin_install (char *name,
		    void (*handler) (eval_environ_t),
		    instr_rettype rettype,
		    size_t argcount,
		    ...)
{
	struct symtab *sp;
	va_list ap;
	size_t i;
	instr_rettype *argtype = emalloc (argcount * sizeof *argtype);

	va_start(ap, argcount);
	for (i = 0; i < argcount; i++) 
		argtype[i] = va_arg(ap, instr_rettype);
	va_end(ap);
	sp = lookup_or_install(SYM_BUILTIN, name, 1);
	if (!sp) {
		parse_error("Cannot install builtin function %s", name);
		abort();
	}

	if (sp->state != SYM_UNDEF) {
		parse_error("INTERNAL ERROR: cannot install builtin function %s", name);
		abort();
	}

	/* new entry */
	sp->state = SYM_BUILTIN;
	sp->v.builtin.name = strdup(name);
	sp->v.builtin.handler = handler;
	sp->v.builtin.parmcount = argcount;
	sp->v.builtin.parmtype = argtype;
	sp->v.builtin.rettype = rettype;
}

const struct builtin *
builtin_lookup(const char *name)
{
	struct symtab *sp = lookup_or_install(SYM_BUILTIN, name, 0);
	return sp ? &sp->v.builtin : NULL;
}



struct variable *
variable_install(const char *name)
{
	struct symtab *sp = lookup_or_install(SYM_VARIABLE, name, 1);
	if (sp->state == SYM_UNDEF) {
		sp->state = SYM_VARIABLE;
		sp->v.variable.name = strdup(name);
		sp->v.variable.type = rettype_unspecified;
		sp->v.variable.off = 0;
	} else if (sp->state != SYM_VARIABLE)
		abort();
	return &sp->v.variable;
}

struct variable *
variable_lookup(const char *name)
{
	struct symtab *sp = lookup_or_install(SYM_VARIABLE, name, 0);
	return sp ? &sp->v.variable : NULL;
}


struct db_format *
db_format_install(struct db_format *fmt)
{
	struct symtab *sp;

	if (!fmt)
		return NULL;
	
	sp = lookup_or_install(SYM_DB_FORMAT, fmt->name, 1);
	sp->state = SYM_DB_FORMAT;
	sp->v.db = *fmt;
	return &sp->v.db;
}

struct db_format *
db_format_lookup(const char *name)
{
	struct symtab *sp = lookup_or_install(SYM_DB_FORMAT, name, 0);
	return sp ? &sp->v.db : NULL;
}


struct function *
function_install(char *name, int parmcnt,
		 instr_rettype *parmtypes, instr_rettype rettype,
		 const struct locus *locus)
{
	struct symtab *sp = lookup_or_install(SYM_FUNC, name, 1);
	if (sp->state == SYM_FUNC) {
		parse_error_locus(locus,
				  "Redefinition of function %s", name);
		parse_error_locus(&sp->v.fun.locus, "Previously defined here");
	} else {
		sp->state = SYM_FUNC;
		sp->v.fun.name = strdup(name);
	}
	sp->v.fun.locus = *locus;
	sp->v.fun.node = NULL;
	sp->v.fun.entry = 0;
	sp->v.fun.parmcount = parmcnt;
	sp->v.fun.parmtype = parmtypes;
	sp->v.fun.rettype = rettype;
	return &sp->v.fun;
}

struct function *
function_lookup(const char *name)
{
	struct symtab *sp = lookup_or_install(SYM_FUNC, name, 0);
	return sp ? &sp->v.fun : NULL;
}
	
